/****************************************************************************
*  pvm.c
*
*  This module implements the PVM control routines for PVM'd POVRAY.
*
*  This file was written by Brad Kline.  April 1994.
*                modified by Andreas Dilger Jan - May 1995
*                modified by Harald Deischinger, 1996 - 1997
*
*  NOTE - PVM is not supplied here.
*
*  You need the PVM 3.x libraries, include files, and PVMD daemons.
*  
*  PVM is available from the University of Tennessee, Knoxville, TN.
*
*  The author disclaims all warranties with regard to this software, 
*  including all implied warranties of merchant-ability and fitness. 
*  The code is simply distributed as it is.
*
* 970610 - New option PVM_WD. To specify a working directory for
*          the slaves. [Deischi]
*          New option PVM_Hosts. To give names for hosts to use
*          as slaves. [Deischi]
*
* 9704.. - changed to to animations, lots of changes
*          simplifications to the whole thing [Deischi]
*
* 96.... - Changed code to work with POVRAY 3.0 [Deischi]
*
* 950508 - Changed code that allocates blocks to be more robust.
*          Changed the write_line routines to hook into the existing routines.
*          Changed the slave startup processes to be more informative.
*
* 940511 - Remove code that tried to fit grid sections into whole parts. The
*          code now supports odd sized images.
*
*****************************************************************************/
#include <sys/time.h>
#include <sys/resource.h>
#include "frame.h"
#include "vector.h"
#include "povproto.h"
#include "frame.h"
#include "povray.h"
#include "render.h"
#include "optout.h"
#include "pvm3.h"
#include "pvm.h"


/* This distribution scheme breaks down the image into to two dimensional
 * areas.  Each area has a designated grid control point.  The grid area
 * is up to PvmChunkWidth (-NWxx) pixels wide by PvmChunkHeight (-NHxx)
 * high. If one uses the default, then a grid section would be 32x32, or
 * 1024 pixels.  In a 640x480 image, this would create 300 grid control
 * points, and therefore 300 different sections to do work on (not lines.).
 */

/*
 * clock value in shellout command is not working correctly
 * timings are not displayed (and done) correctly
 * statistics can be wrong
 * continue trace might produce wrong results -- needs testing
 */

/*
 * Global variables
 */
char    PvmArch[20] = { 0 };
int     PvmChunkHeight = PVM_DEFAULT_GRID_HEIGHT;	/* height of a block */
int     PvmChunkWidth  = PVM_DEFAULT_GRID_WIDTH;	/* widht of a block */
int     PvmRowsToSend = 1;
int     PvmSlave = 0;
int     PvmTasks = 9999;
int     PvmNice = PVM_DEFAULT_NICE;
int     PvmMTid;
int     PvmBlockNum;
char *  PvmSlavename = NULL;
int	PvmNextFrame = 0;
char	PvmWorkingDir[PATH_MAX+1] = "";			/* PVM Working Directory */
char ** PvmHosts = NULL;
int	PvmHostsN = 0;

/*
 * local functions
 */
static void PvmSpawnError PARAMS((int));
static void PvmSpawnTidError PARAMS((int));
static void PvmSpawn PARAMS((void));
static void PvmCheckFirst PARAMS((void));
static void PvmInitSlaveStat PARAMS((void));
static void PvmInitFrameStat PARAMS((void));

static int PvmMasterReceive PARAMS((void));
static void PvmIdentifySlave PARAMS((void));
static void PvmSendWork PARAMS((int block));
static void PvmReassignWork PARAMS((int block));
static void PvmAssignWork PARAMS((void));
static int PvmStartFrame PARAMS((void));
static void PvmFinishFrame PARAMS((void));
static void PvmReceiveData PARAMS((void));
static void PvmWrite PARAMS((void));

static void PvmUnpackStats PARAMS((void));
static void PvmUnpackErrorReport PARAMS((void));


static int grid_points;		/* number of grid blocks for one frame */
static int grid_ppline;		/* Number of grid blocks in a row */
static int frames;		/* num. of frames to do */
static int frame = -1;		/* the current frame, start with 0 to count */
static int next_frame_finish;	
static int frame_display;	/* the frame that is displayed */

static int grid_lines_done = 0;	/* num. of PvmCthunkWidth lines done (only statistics) */

static int npixels_done = 0;    /* Number of pixels done, by all processors */

static int                row_start;           /* First row in the block */
static int                row_end;             /* Last row in the block */
static int                col_start;           /* First column in the block */
static int                col_end;             /* Last column in the block */
static int msgtag; 

static   struct timeval     wait_time;           /* Time to wait for slave stats */
static   struct pvmhostinfo *hostinfo;           /* Info on each PVM host */
static   int                minions;             /* Number of slave tasks */
static   int                nhost;               /* Number of available pvm hosts */
static   pvm_slave_stat     *slave_stat;         /* Info on the slaves */
static   pvm_frame_stat     *frame_stat;
static   int                rtid;                /* tid of the message sender */
static   int                *tids;               /* tids of all the slaves */
static   int                slave_num;           /* Relative host number of sender */
static   char               *slave_name;         /* Hostname of sending slave */
static   int                nstat = 0;           /* Number of slaves returning stats */

static   int                bitmask;             /* Start tasks on a specific arch */
static   int                bufid;               /* pvm recieve buffer id */
static   int                bytesin;             /* Size of message */
static   char               **sargs;             /* argv for the slaves */
static   char               slave_option[PATH_MAX+4]; /* the directory argument */


/* in povray.c */
extern char Actual_Output_Name[FILE_NAME_LENGTH];
void setup_output_file_name PARAMS((void));
void open_output_file PARAMS((void));
void set_output_file_handle();



/********************************************************************************
 * Initialize Master Process
 *******************************************************************************/


/*
 * Spawn the slaves, and check for errors 
 */

/* handle an error from pvm_spawn */
static void PvmSpawnError(m) 
    int m; 
{
    /* Only quit if we can't start many tasks, or an error condition. */
    Error_Line( "...error spawning tasks because \n");
    switch (m) {
    case PvmBadParam:
	Error_Line("of bad parameter.\n");
	break;
    case PvmNoHost:
	Error_Line("host not in PVM.\n");
	break;
    case PvmNoFile:
	Error_Line("executable not found.\n");
	break;
    case PvmNoMem:
	Error_Line("no memory on host.\n");
	break;
    case PvmSysErr:
	Error_Line("pvmd not responding.\n");
	break;
    case PvmOutOfRes:
	Error_Line("lack of resources.\n");
	break;
    default:
	Error_Line("of unknown error.\n");
    }
    pvm_perror("");
    pvm_exit();
    close_all();
    exit(1);
}

static void PvmSpawnTidError(tid) 
    int tid; 
{

    Warning(0.0, "...spawn failure because ");
    switch (tid) {
    case PvmBadParam:
	Warning(0.0, "of bad parameter.\n");
	break;
    case PvmNoHost:
	Warning(0.0, "host not in PVM.\n");
	break;
    case PvmNoFile:
	Warning(0.0, "executable not found.\n");
	break;
    case PvmNoMem:
	Warning(0.0, "no memory on host.\n");
	break;
    case PvmSysErr:
	Warning(0.0, "pvmd not responding.\n");
	break;
    case PvmOutOfRes:
	Warning(0.0, "lack of resources.\n");
	break;
    default:
	Warning(0.0, "of unknown error.\n");
    }
}

static void PvmSpawn() {
    int i;

    /* Start up the slave processes */
    if (PvmArch[0] != '\0') {
	Status_Info("  Spawning %s with %d PVM tasks on arch %s...\n",
		    PvmSlavename, PvmTasks,PvmArch);
    } else {
	Status_Info("  Spawning %s with %d PVM tasks on %d hosts...\n",
		    PvmSlavename, PvmTasks,nhost);
    }

    if(PvmHosts) {			/* start at specified hosts */
	int m;
	minions = 0;
	for(i=0; i < PvmTasks; i++) {	/* start one by one */
	    m = pvm_spawn(PvmSlavename, sargs, PvmTaskHost, PvmHosts[i % PvmHostsN], 
			  1, tids+minions);
	    if( m < 0) {			/* spawn failed */
		PvmSpawnError(m);
	    } else if (m==0) {			/* failed a little bit */
		PvmSpawnTidError(tids[minions]);
	    }
	    minions += m;
	}

    } else {				/* start all at once somwhere */
	minions = pvm_spawn(PvmSlavename, sargs, bitmask, PvmArch, PvmTasks, tids);

	if(minions < 0)			/* check for total failure */
	    PvmSpawnError(minions);
	for (i = minions; i < PvmTasks; i++)	/* errors/process */
	    PvmSpawnTidError(tids[i]);
    }

    /* Check if "few" tasks could be started */
    if (minions < (PvmTasks + 1) / 2)  {
	if(minions==0) {
	    Error_Line("...No tasks");
        } else {
	    Error_Line("...only %d of %d tasks", minions, PvmTasks);
	}
	Error_Line(" spawned!  Quitting.\n");
	pvm_exit();
	close_all();
	exit(1);
    } else {
	Status_Info("  ...%d PVM tasks successfully spawned.\n", minions);
    }

    PvmTasks = minions;
}


/*
 * Wait up to 120 seconds for the first to check in
 */
static void PvmCheckFirst() {
    wait_time.tv_sec = 120;
    wait_time.tv_usec = 0;

    Status_Info("  Waiting up to %lds for first slave to start...\n",
		wait_time.tv_sec);
    minions = 0;

    if( pvm_trecv(-1, PVM_INIT_SLAVE, &wait_time) <= 0 ) {
	/* No slave checked in. Something is wrong */
	Error_Line("First slave did not start in time!  Quitting.\n");
	Error_Line("See /tmp/pvml.<your uid> on the master PVM host for\n");
	Error_Line("more info on why the slave didn't start correctly.\n");
	pvm_exit();
	close_all();
	exit(1);
    }  else {
	minions++;
    }

    if (opts.Options & VERBOSE) {
	Status_Info("  Slave 0 successfully started.\n");
    }
}


/* 
 * Get a status struct for each slave 
 */
static void PvmInitSlaveStat() {
    
    slave_stat = (pvm_slave_stat *)POV_CALLOC(PvmTasks, sizeof(pvm_slave_stat), "slave_stat");
    
    /* Set up pointers to all the host names and the slave status */
    for (slave_num = 0; slave_num < PvmTasks; slave_num++) {
	int j;
	int dtid;

	dtid = pvm_tidtohost(tids[slave_num]);
	slave_stat[slave_num].name = "unknown";

	for (j = 0; j < nhost; j++) {
	    if (dtid == hostinfo[j].hi_tid) {
		slave_stat[slave_num].name = (char*)POV_CALLOC(strlen(hostinfo[j].hi_name)+1, 
							       sizeof(char), "hostnames");
		strcpy(slave_stat[slave_num].name,hostinfo[j].hi_name);
		break;
	    }
	}
	slave_stat[slave_num].frame_assigned = PVM_SLAVE_FREE;
	slave_stat[slave_num].block_reassigned = PVM_SLAVE_FREE;
    }
}

/* 
 * Get a status struct for each frame
 */
static void PvmInitFrameStat() {
    int i,j;

    frames = (opts.FrameSeq.FrameType == FT_MULTIPLE_FRAME) ? 
	opts.FrameSeq.FinalFrame - opts.FrameSeq.InitialFrame + 1 :
	1;
    frame_stat = (pvm_frame_stat *)POV_CALLOC(frames, sizeof(pvm_frame_stat), "frame_stat");
    frame_display = -1;

    for(i=0; i < frames; i++) {
	frame_stat[i].block_to_assign = 0;
	frame_stat[i].block_to_reassign = 0;

	frame_stat[i].block_to_write = 0;

	frame_stat[i].row_to_write  = opts.First_Line;
	frame_stat[i].row_to_malloc = opts.First_Line;

	frame_stat[i].block_stat = 
	    (pvm_block_stat *)POV_CALLOC(grid_points, sizeof(pvm_block_stat), "block_stat");

	for(j=0; j < MaxStat; j++)
	    Init_Counter(frame_stat[i].stats[j]);
    }
}

/*
 * Initialize PVM Master
 */
void 
Pvm_Master_Init(argc, argv)
    int     argc;
    char    **argv;
{
    int                i;

    Status_Info("Initializing PVMPOV\n");

    /* Get a table for the TIDs */
    tids = (int *)POV_CALLOC(PvmTasks, sizeof(int), "tids");

    /* Build a new arg list for the sub-tasks - add -ni in front */
    sargs = (char **)POV_CALLOC(argc + 1, sizeof(char *), "sargs");

    /* Move argument pointers to new list array, keep position 0 free for -ni option */
    for (i = 1; i < argc; i++) {
	sargs[i] = argv[i];
    }

    /* Get current directory for the slave tasks */
    if(PvmWorkingDir[0] == '\0')
	if ( GETCWD(PvmWorkingDir) == NULL) {
	    Error("PVM Master cannot get current directory.\n");
	    return;
	}

    /* Add slave flag */
    sprintf(slave_option, "-nI%s", PvmWorkingDir);
    sargs[0] = slave_option;

    /* Enroll in PVM */
    if (pvm_mytid() < 0) {
	Error("Error starting master task.\n");
    }

    if (pvm_config(&nhost, (int *)NULL, &hostinfo) < 0) {
	pvm_exit();
	Error("Error getting PVM machine status.\n");
    }
           
    bitmask = PvmTaskDefault;

    /* If PvmArch is specified, use that architecture */
    if (PvmArch[0] != '\0') {
	bitmask = PvmTaskArch;

	/* If arch is psecified, but not the number of tasks, go and
	 * count how many hosts of that type there are.
	 */
	if (PvmTasks >= 9999) {
	    PvmTasks = 0;
	    
	    for (i = 0; i < nhost; i++) {
		if (strcmp(PvmArch,hostinfo[i].hi_arch) == 0)
		    PvmTasks++;
	    }
	}    
    }

    /* If PvmTasks is specified, use that many tasks, otherwise use
     * as many hosts as possible.
     */
    if (PvmTasks >= 9999) {
	PvmTasks = nhost;
    }

    /* if no slavename was given, use name of current executable */
    if(PvmSlavename == NULL)
	PvmSlavename = argv[0];

    PvmSpawn();						/* spawn the slaves */
    PvmCheckFirst();					/* wait for first slave */

    pvm_setopt(PvmRoute, PvmRouteDirect);

    /* Max # of pixels in a grid row */
    PvmChunkWidth = min(PvmChunkWidth, opts.Last_Column - opts.First_Column);

    /* Grid points per line - rounded up */
    grid_ppline = (opts.Last_Column  - opts.First_Column + PvmChunkWidth - 1) / PvmChunkWidth;

    /* Establish grid section height */
    PvmChunkHeight = min(PvmChunkHeight, opts.Last_Line - opts.First_Line);

    /* Total grid points in the image - round up PvmChunkHeight */
    grid_points = ((opts.Last_Line - opts.First_Line + PvmChunkHeight - 1) / PvmChunkHeight) * 
	grid_ppline;
    
    PvmInitSlaveStat();
    PvmInitFrameStat();

    /* initialize total stats */
    for(i=0; i < MaxStat; i++) {
	Init_Counter(totalstats[i]);
    }
}





/********************************************************************************
 * Master Control
 ********************************************************************************/

/*
 * Receive a message
 */
static int PvmMasterReceive() {

    /* wait for a message from the slaves */
    if ((bufid = pvm_trecv(-1, -1, &wait_time)) < 0) {
	Error_Line( "PVM Master cannot receive.\n");
	pvm_perror("");
	pvm_exit();
	exit(1);
    }

    /* We timed out waiting for any messages */
    if (bufid == 0) {
	Status_Info(".");
	fflush(stderr);
	
	return 1;
    }
    
    return 0;
}

/*
 * Find name and number of slave, of current message
 */
static void PvmIdentifySlave() {

    /* get info about this msg. */
    if (pvm_bufinfo(bufid, &bytesin, &msgtag, &rtid) < 0) {
	Error_Line( "PVM Master cannot bufinfo.\n");
	pvm_perror("");
	pvm_exit();
	return;
    }

    /* find the name of the slave, sending the msg. */
    slave_name = "unknown";
    for (slave_num = 0; slave_num < PvmTasks; slave_num++) {
	if (tids[slave_num] == rtid) {
	    slave_name = slave_stat[slave_num].name;
	    break;
	}
    }
}

/*
 * Send a PVM_WORK message
 * Set block to PVM_GRID_POINT_ASSIGNED
 */
static void PvmSendWork(block) 
    int block;
{
    int row_start, row_end, col_start, col_end;
#ifdef DEBUG
    Status_Info("Assigning block: %d, frame: %d of %d to %s.\n", block, frame, frames, slave_name);
#endif /* DEBUG */

    frame_stat[frame].block_stat[block].status = PVM_GRID_POINT_ASSIGNED;
    frame_stat[frame].block_stat[block].assigned_tid = rtid;
    
    row_start = (block/grid_ppline) * PvmChunkHeight + opts.First_Line;
    row_end   = (row_start + PvmChunkHeight) - 1;
	    
    if (row_end >= opts.Last_Line)
	row_end = opts.Last_Line - 1;
    
    col_start = (block%grid_ppline) * PvmChunkWidth + opts.First_Column;
    col_end   = (col_start + PvmChunkWidth) - 1;
    
    if (col_end >= opts.Last_Column)
	col_end = opts.Last_Column - 1;
    
    pvm_initsend(PvmDataDefault);
    pvm_pkint(&frame,1,1);
    pvm_pkint(&block,1,1);
    pvm_pkint(&row_start,1,1);
    pvm_pkint(&row_end,1,1);
    pvm_pkint(&col_start,1,1);
    pvm_pkint(&col_end,1,1);
    if (pvm_send(rtid, PVM_WORK) < 0) {
	Error_Line( "PVM Master can't send to slave on %s.\n",slave_name);
	pvm_perror("");
	pvm_exit();
	exit(1);
    }

    frame_stat[frame].block_to_assign ++;
}

/*
 * Send a PVM_WORK message
 * Set block to PVM_GRID_POINT_REASSIGNED
 */
static void PvmReassignWork(block) 
    int block;
{
#ifdef DEBUG
    Status_Info("Reassigning %d to %s.\n", block,
		slave_name);
#endif /* DEBUG */
    
    frame_stat[frame].block_stat[block].status=PVM_GRID_POINT_REASSIGNED;
    frame_stat[frame].block_stat[block].reassigned_tid = rtid;
    slave_stat[slave_num].block_reassigned = block;
			
    if (frame_stat[frame].block_stat[block].wr_line == 0) {
	row_start = (block / grid_ppline)*PvmChunkHeight +  opts.First_Line;
    } else {
	row_start = frame_stat[frame].block_stat[block].wr_line;
    }
    row_end = (block/grid_ppline + 1) * PvmChunkHeight - 1;
    
    if (row_end >= opts.Last_Line)
	row_end = opts.Last_Line - 1;
			
    col_start = (block % grid_ppline) * PvmChunkWidth +
	opts.First_Column;
    col_end   = (col_start + PvmChunkWidth) - 1;
    
    if (col_end >= opts.Last_Column)
	col_end = opts.Last_Column - 1;
    
    pvm_initsend(PvmDataDefault);
    pvm_pkint(&frame,1,1);
    pvm_pkint(&block,1,1);
    pvm_pkint(&row_start,1,1);
    pvm_pkint(&row_end,1,1);
    pvm_pkint(&col_start,1,1);
    pvm_pkint(&col_end,1,1);
    if (pvm_send(rtid, PVM_WORK) < 0) {
	Error_Line( "PVM Master can't send to slave on %s.\n",slave_name);
	pvm_perror("");
	pvm_exit();
	exit(1);
    }

    frame_stat[frame].block_to_reassign++;
}


/*	 
 * Do all the stuff, that is necessary for a new frame 
 * This is called before the PVM_WORK is sent
 *
 * A lot of this is taken from 'FrameRender' 
 *
 * ret: 0 .. OK send the PVMWOR 
 *      1 .. Shellout gave SKIP_ONCE_RET -> this frame is done
 */
static int PvmStartFrame() {
    int Frame_Result;

    if(frame_stat[frame].status != PVM_FRAME_FREE) {
	Warning(0.0, "Frame %d already started.\n", frame);
	return 1;
    }

    if(opts.Options & VERBOSE)
	Status_Info("\nStarting frame %d...", opts.FrameSeq.InitialFrame + frame);
    
    frame_stat[frame].status = PVM_FRAME_WORKING;

    /* allocate row pointers */
    frame_stat[frame].row_ptr = (COLOUR **)POV_CALLOC(opts.Last_Line, sizeof(COLOUR *), "row_ptr");

    opts.FrameSeq.FrameNumber = opts.FrameSeq.InitialFrame + frame;
    setup_output_file_name();
    
    /* Execute a shell-out command before tracing */
    Frame_Result = POV_SHELLOUT(PRE_FRAME_SHL);
    
    if (Frame_Result == ALL_SKIP_RET) {
	/* stop completely */
	int i;
	for(i=frame; i < frames; i++) {
	    frame_stat[i].block_to_assign = grid_points;
	    frame_stat[i].block_to_reassign = grid_points;
	    frame_stat[i].status = PVM_FRAME_DONE_SHELLOUT;
	}
	/* continue searching, frames before this one, should be finished */
	return 1;		
    }
    
    if (Frame_Result == SKIP_ONCE_RET) {
	/* mark frame as done */
	frame_stat[frame].block_to_assign = grid_points;
	frame_stat[frame].block_to_reassign = grid_points;
	frame_stat[frame].status = PVM_FRAME_DONE_SHELLOUT;

	/* continue searching */
	return 1;
    }

    /*
     * open output file
     */
    set_output_file_handle();
    open_output_file();

    Initialize_Renderer(); 

    /*
     * Start displaying
     */
    if( (opts.Options & DISPLAY) && !Display_Started)  {
	Status_Info("\nDisplaying frame %d...", opts.FrameSeq.InitialFrame + frame);
	POV_DISPLAY_INIT(Frame.Screen_Width, Frame.Screen_Height);
	Display_Started = TRUE;
	frame_display = frame;
	Status_Info("\n");
    }
    
    
    /*
     * read already rendered parts
     */
    if ((opts.Options & DISKWRITE) && (opts.Options & CONTINUE_TRACE)) {
	int i;
	int fl = opts.First_Line;			/* remember start of image */
	
	Read_Rendered_Part(Actual_Output_Name);		/* read -> changes opts.First_line */

	frame_stat[frame].row_to_write = 
	    frame_stat[frame].row_to_malloc = opts.First_Line;

	frame_stat[frame].block_to_write = 
	    frame_stat[frame].block_to_assign = (opts.First_Line/PvmChunkHeight)*grid_ppline;
	/* blocks above the new start are already finished */
	for(i=0; i < frame_stat[frame].block_to_assign; i++) {
	    frame_stat[frame].block_stat[i].status = PVM_GRID_POINT_WRITTEN;
	}
	if(opts.Options & VERBOSE)
	    Status_Info("Continuing with block %d in frame %d\n", 
			frame_stat[frame].block_to_assign, frame);

	opts.First_Line = fl;				/* set back the old value */


	/* this should not be necessary */
	if (opts.Last_Line > Frame.Screen_Height)
	    opts.Last_Line = Frame.Screen_Height;
	
	if (opts.Last_Column > Frame.Screen_Width)
	    opts.Last_Column = Frame.Screen_Width;	
    }

    /*
     *  make a copy of the Output_File_Handle 
     */
    frame_stat[frame].output_file = Output_File_Handle;
    if(Output_File_Handle->filename) {
	char * n = POV_MALLOC(strlen(Output_File_Handle->filename)+1, "filename");
	strcpy(n, Output_File_Handle->filename);
	frame_stat[frame].output_file->filename = n;
    }

    if(opts.Options & VERBOSE)
	Status_Info("\n");

    return 0;
}


/*
 * test, if the next frame in the sequence finished,
 * do all the necessary stuff to close the frame (shellout, ...)
 */
static void PvmFinishFrame() {

    /* check, if frame has really finished */
    if( frame_stat[frame].row_to_write < opts.Last_Line)
	return;

    /* if not already closed */
    if( frame_stat[frame].status < PVM_FRAME_DONE) {
	
	if(opts.Options & VERBOSE)
	    Status_Info("\nFinishing frame %d...", opts.FrameSeq.InitialFrame + frame);
	Status_Info("rtw. %d\n", frame_stat[frame].row_to_write);

	frame_stat[frame].status = PVM_FRAME_DONE;
	
	POV_FREE(frame_stat[frame].row_ptr);
	
	/* Close output file */
	if (frame_stat[frame].output_file) {
	    Output_File_Handle = frame_stat[frame].output_file;
	    Close_File(frame_stat[frame].output_file);
	}
	
	if( frame == frame_display) {
	    POV_DISPLAY_FINISHED
		
		if ((opts.Options & DISPLAY) && Display_Started)  {
		    POV_DISPLAY_CLOSE
	    
			Display_Started = FALSE;
		}
	    
	    frame_display = -1;
	}
	
	if (opts.histogram_on)
	    write_histogram (opts.Histogram_File_Name);
    }

    if( frame_stat[frame].status < PVM_FRAME_DONE_SHELLOUT) {
	/* the shellouts are done in the correct order */
	if( frame == next_frame_finish) {
	    int Frame_Result;

	    frame_stat[frame].status = PVM_FRAME_DONE_SHELLOUT;

	    opts.FrameSeq.FrameNumber = opts.FrameSeq.InitialFrame + frame;
	    setup_output_file_name();
	    
	    Frame_Result = POV_SHELLOUT(POST_FRAME_SHL);
	    
	    if ((Frame_Result==SKIP_ONCE_RET) || (Frame_Result==ALL_SKIP_RET)) {
		next_frame_finish = frames;
		return;
	    }
	    next_frame_finish ++;	/* one more frame finished */
	    
	    frame ++;		/* check next frame (might have finished earlier) */
	    if( frame < frames)
		PvmFinishFrame();
	}
    }
    if(opts.Options & VERBOSE)
	Status_Info("\n");	
}


/*
 * Assign Work
 *
 * A slaved requested work, now find free block (or a block to reassign)
 */
static void PvmAssignWork() {
    int min_assigned;
    int found;
    int f;
    
    /* The slave wants work, but has already been reassigned something,
     * then tell it to ask again when it really wants work.
     */
    if( slave_stat[slave_num].block_reassigned != PVM_SLAVE_FREE) {
#ifdef DEBUG
	Status_Info("%s already assigned on %d.  Not reassigning.\n",
		    slave_name,
		    slave_stat[slave_num].block_reassigned);
#endif /* DEBUG */
	pvm_initsend(PvmDataDefault);
	pvm_send(rtid, PVM_INIT_SLAVE);
	
	return;
    }


    /*
     * search for a free block, take a frame with minimal
     * number of assigned blocks
     */
    found = 0;
    min_assigned = grid_points;

    frame = slave_stat[slave_num].frame_assigned;
    /* if slave was already assigned and that frame has still free blocks */
    if ( (frame >= 0) &&
	 (frame_stat[frame].block_to_assign < grid_points) ) {
	
	min_assigned = 0;
	found = 1;
    }
    
    /* find free frame with minimal number of assigned frames */
    for(f = 0; (f < frames) && (min_assigned > 0); f ++) {
	if( (frame_stat[f].block_to_assign < min_assigned) &&
	    (frame_stat[f].block_to_assign < grid_points) ) {
	    
#ifdef DEBUG
	    Status_Info("\nfound: frame %d, ass: %d, min: %d\n", 
			f, frame_stat[f].block_to_assign,
			min_assigned);
#endif

	    min_assigned = frame_stat[f].block_to_assign;
	    frame = f;
	    found = 1;

	}
    }

    if( found ) {
	/* first block for that frame */
	if( frame_stat[frame].block_to_assign == 0) {	
	    /* Do the necessary preparations for a new frame */
	    if( PvmStartFrame() ) {
		/* Shellout gave SKIP_ONCE_REt -> search next block */
		PvmAssignWork();
		return;
	    }
	}

	slave_stat[slave_num].frame_assigned = frame;
	PvmSendWork(frame_stat[frame].block_to_assign);
	return;
    }


    /*
     * found no free block
     * search for a block to reassign
     */

    /* if this slave was slow, and there are enough other slaves left,
       don't use this anymore */
    if ( (slave_stat[slave_num].pixels_done < npixels_done*9/10/PvmTasks) &&
	 (minions >= (PvmTasks + 3) / 4) ) {
    
	Status_Info("Not using %s for reassignment (%d%%)\n", slave_name,
		    100*slave_stat[slave_num].pixels_done*PvmTasks/npixels_done);

	/* tell slave to finish work */
	pvm_initsend(PvmDataDefault);
	pvm_send(rtid, PVM_STOP_SLAVE);
    }

    found = 0;

    frame = slave_stat[slave_num].frame_assigned;
    /* if slave was already assigned and that frame has free blocks */
    if ( (frame >= 0) &&
	 (frame_stat[frame].block_to_reassign < grid_points) ) {
	
	min_assigned = 0;
	found = 1;
    }
	
    /* find first frame with minimal number of reassigned frames */
    min_assigned = grid_points + 1;
    for(f = 0; (f < frames) && (min_assigned > 0); f ++) {
	if( frame_stat[f].block_to_reassign < min_assigned) {
	    
	    min_assigned = frame_stat[frame].block_to_reassign;
	    frame = f;
	    found = 1;
	}
    }
    
    if( found ) {
	/* find block in frame */
	int block;
	for(block = 0; block < grid_points; block ++) {
	    if( frame_stat[frame].block_stat[block].status < PVM_GRID_POINT_DONE) {
		slave_stat[slave_num].frame_reassigned = frame;
		PvmReassignWork(block);
		return;
	    }
	}
    }

    /* 
     * there is nothing left to reassign
     */
    Status_Info("All blocks are assigned. Stopping %s.\n", slave_name);

    /* tell slave to finish work */
    pvm_initsend(PvmDataDefault);
    pvm_send(rtid, PVM_STOP_SLAVE);

}

/*
 * display a part from one line
 */
static void PvmDisplayPlot(row_ptr, row, col, ncols) 
    COLOUR * row_ptr;
    int row, col, ncols;
{
    unsigned char Red, Green, Blue, Alpha;
    DBL grey;
    int i;
    
    /* draw line */
    for (i = col; i < col+ncols; i++) {
	extract_colors(row_ptr[i], &Red, &Green, &Blue, &Alpha, &grey);
	POV_DISPLAY_PLOT(i, row, Red, Green, Blue, Alpha);
    }
    /* force draw of the whole line (with X11) */
    i = opts.Last_Column - 1;
    extract_colors(row_ptr[i], &Red, &Green, &Blue, &Alpha, &grey);
    POV_DISPLAY_PLOT(i, row, Red, Green, Blue, Alpha);
}


/*
 * one of the slaves sends a part of it's computed data
 */
static void PvmReceiveData() {
    int ncols;
    int i;
    int block;
    
    pvm_upkint(&frame, 1, 1);
    pvm_upkint(&block, 1, 1);
    pvm_upkint(&row_start, 1, 1);
    pvm_upkint(&col_start, 1, 1);
    pvm_upkint(&ncols, 1, 1);

    row_end = row_start + (bytesin - 3*sizeof(int))/(5*ncols*sizeof(COLC));
    col_end = col_start + ncols;

#ifdef DEBUG
    Status_Info("receiving frame: %d, row: %d-%d, col: %d-%d\n", 
		frame, row_start, row_end, col_start, col_end);
#endif

    /* Make some checks to see if the data is not corrupted or invalid */
    if( (row_start < opts.First_Line) || (col_start < opts.First_Column) ||
	(row_end   > opts.Last_Line)  || (col_end   > opts.Last_Column) ||
	(frame     < 0)               || (frame     >= frames) ||
	(ncols     > PvmChunkWidth)   || (row_end - row_start > PvmChunkHeight) ) {
	
	Warning(0.0, "Bad block recieved from %s\n",slave_name);
	Warning(0.0, "Frame %d, Row %d - %d, Column %d - %d.\n", frame, row_start, row_end,col_start, col_end);
	
	return;
    }
    
    /* get one row after the other */
    for (i = row_start; i < row_end; i++) {
	if (i >= frame_stat[frame].block_stat[block].wr_line) {  /* Is incoming data current? */

	    /* allocate memory, to receive data */
	    if (frame_stat[frame].row_ptr[i] == NULL) { /* Starting a new row */
		/* FD 00-10-20: modified the allocated buffer size to the whole picture width. Fixes  bug crashing pvmpov when using the -ec flag */
		frame_stat[frame].row_ptr[i] = (COLOUR *)POV_MALLOC(Frame.Screen_Width*sizeof(COLOUR), 
								    "row_ptr");
		
		while(frame_stat[frame].row_ptr[frame_stat[frame].row_to_malloc] != NULL && 
		      frame_stat[frame].row_to_malloc < opts.Last_Line) {
		    frame_stat[frame].row_to_malloc++;
		}
	    }
	    
	    /* get color data */
	    pvm_upkCOL(&(frame_stat[frame].row_ptr[i][col_start][0]),5*ncols,1);
	    
	    frame_stat[frame].block_stat[block].wr_line = i + 1;

	    if (frame == frame_display) {
		PvmDisplayPlot(frame_stat[frame].row_ptr[i], i, col_start, ncols);
	    }

	    /* Statistics - Attribute this grid section  */
	    grid_lines_done++;
	    slave_stat[slave_num].pixels_done += ncols;

	    npixels_done += ncols;
	} else {  
	    /* another slave already sent this data */

	    COLOUR *dummy = (COLOUR *)POV_MALLOC(ncols * sizeof(COLOUR), "dummy row");
	    pvm_upkCOL(&(dummy[0][0]),5*ncols,1);
	    
	    /* Statistics - Attribute this grid section  */
	    slave_stat[slave_num].pixels_late += ncols;
	    POV_FREE(dummy);
	}
    }
    /* Has this entire block been rendered? */
    if (frame_stat[frame].block_stat[block].wr_line >= 
	(block/grid_ppline + 1)*PvmChunkHeight + opts.First_Line - 1 ||
	frame_stat[frame].block_stat[block].wr_line >= opts.Last_Line)  {
#ifdef DEBUG
	Status_Info("Block completed by %s\n", slave_name);
#endif /* DEBUG */
	frame_stat[frame].block_stat[block].status = PVM_GRID_POINT_DONE;
	if (slave_stat[slave_num].block_reassigned == block) {
#ifdef DEBUG
		Status_Info("   reassigned block cleared.\n");
#endif /* DEBUG */
		slave_stat[slave_num].block_reassigned = PVM_SLAVE_FREE;
	}
#ifdef DEBUG
	else
	{
	    Status_Info("   reassigned block is %d.\n",
			slave_stat[slave_num].block_reassigned);
	}
#endif /* DEBUG */
    }
    
    if ((opts.Options & VERBOSE) && (grid_lines_done % PvmChunkHeight) == 0) {
      DBL time_dif;
      unsigned long hrs, mins;
      DBL secs;
      
      STOP_TIME
	
	time_dif = TIME_ELAPSED
	
	if (time_dif > 0.0)
	  {
	    SPLIT_TIME(time_dif, &hrs, &mins, &secs);
	    
	    Status_Info("\r%3ld:%02ld:%02ld ", hrs, mins, (long)secs);
	  }
	else
	  {
	    Status_Info("\r  -:--:-- ");
	  }
      Status_Info("%5.2f of blocks complete.",
		  (float)100*grid_lines_done/PvmChunkHeight/grid_points/frames);
    }
}

/*
 * Write out us many lines as possible 
 */
static void PvmWrite() {
    int i;

    /* FD 00-10-20: prevents writing lines "invented" when late
       reattributed blocks are received after the file is closed,
       crashing the master. Seems to fix the segfault happening with
       very fast renderings. */
    if (frame_stat[frame].block_to_write >= grid_points) return;

    do { 
	/* check, if there is a complete line to write */
	for (i = frame_stat[frame].block_to_write; 
	     i < frame_stat[frame].block_to_write + grid_ppline; i++)	{
	    if (frame_stat[frame].block_stat[i].wr_line <= frame_stat[frame].row_to_write) {
#ifdef DEBUG
		Status_Info("needing block %d in frame %d\n", i, frame); 
#endif
		return;
	    }
	}

	/* OK, we can write a whole line */
	if (opts.Options & DISKWRITE) {
#ifdef DEBUG
	    Status_Info("write: %d %d\n", frame, frame_stat[frame].row_to_write);
#endif
	    Write_Line(frame_stat[frame].output_file,
		       frame_stat[frame].row_ptr[frame_stat[frame].row_to_write],
		       frame_stat[frame].row_to_write);
	}

	/* Still lines to allocate so */
	if (frame_stat[frame].row_to_malloc < opts.Last_Line)  {
	    /* move this row buffer down, so we avaoid one malloc/free  */
	    frame_stat[frame].row_ptr[frame_stat[frame].row_to_malloc] = 
		frame_stat[frame].row_ptr[frame_stat[frame].row_to_write];
	    frame_stat[frame].row_ptr[frame_stat[frame].row_to_write] = NULL;
	    frame_stat[frame].row_to_malloc++;
	} else {                        /* Otherwise free the row buffer */
	    POV_FREE(frame_stat[frame].row_ptr[frame_stat[frame].row_to_write]);
	    frame_stat[frame].row_ptr[frame_stat[frame].row_to_write]= NULL;
	}
	
	frame_stat[frame].row_to_write++;
	
#ifndef DEBUG
	if ((opts.Options & VERBOSE) && (frame_stat[frame].row_to_write % 8) == 0) {
	  DBL time_dif;
	  unsigned long hrs, mins;
	  DBL secs;
	  
	  STOP_TIME
	    
	    time_dif = TIME_ELAPSED
	    
	    if (time_dif > 0.0)
	      {
		SPLIT_TIME(time_dif, &hrs, &mins, &secs);
		
		Status_Info("\r%3ld:%02ld:%02ld ", hrs, mins, (long)secs);
	      }
	    else
	      {
		Status_Info("\r  -:--:-- ");
	      }
	  Status_Info("%5.2f of blocks complete.",
		      (float)100*grid_lines_done/PvmChunkHeight/grid_points/frames);
	  Status_Info("  %4d of %4d lines finished (in frame %d).",
		      frame_stat[frame].row_to_write - opts.First_Line, 
		      opts.Last_Line - opts.First_Line, frame);
	}	
#endif /* not DEBUG */
	
	/* Finished a whole row  of blocks */
	if ((frame_stat[frame].row_to_write - opts.First_Line) % PvmChunkHeight == 0) {
	    for (i = frame_stat[frame].block_to_write; 
		 i < frame_stat[frame].block_to_write + grid_ppline; i++) {
		frame_stat[frame].block_stat[i].status = PVM_GRID_POINT_WRITTEN;
	    }
	    frame_stat[frame].block_to_write += grid_ppline;
	}
	
    } while(frame_stat[frame].block_to_write < grid_points);
}

/*
 * Handle slave requests, distribute and collect work
 */
void Pvm_Master_Control() {

    /* Wait five minutes at a time between messages */
    wait_time.tv_sec = 1;/* FD 00-10-28 was 300; improves SIGxxx reactivity */

    next_frame_finish = 0;		/* next frame that finishes */

    /* There are incomplete frames and/or incoming messages, and there are any slaves left*/
    while ( (next_frame_finish < frames) || 
	    ((pvm_probe(-1, -1) > 0) && (minions > 0))) {

	/* check for user abort */
	TEST_ABORT;
	if (Stop_Flag)  {
	    Render_Info("\nAborting render...\n");
	    
	    PvmMasterEnd();

	    if ((opts.Options & DISPLAY) && Display_Started) {
		POV_DISPLAY_CLOSE;
	    }	    
	    if (opts.Do_Stats) {
		PRINT_STATS(stats);
	    }
	    Error("User abort.\n");

	    return;
	}

	/* receive a request from a slave */
	if(PvmMasterReceive())
	    continue;

	PvmIdentifySlave();

#ifdef DEBUG
	Status_Info("Received msg %d from %s\n", msgtag, slave_name);
#endif

	/* a slaves requests work */
	switch(msgtag) {
	case PVM_INIT_SLAVE:		/* Another slave started ok */
	    if (opts.Options & VERBOSE) {
		Status_Info(" Slave %d at %s successfully started.\n", minions, slave_name);
	    }
	    minions++;
	    break;

	case PVM_NEED_WORK:		/* a slave wants work */
	    PvmAssignWork();
 	    break;
	    
	case PVM_SLAVE_RESULTS:		/* slaves sends us some data */
	    PvmReceiveData();
	    PvmWrite();
	    PvmFinishFrame();
	    break;
	    
	case PVM_SLAVE_STATS:		/* Too many stats, eh? */
#ifdef DEBUG
	    Status_Info("Getting stats from %s.\n", slave_name);
#endif /* DEBUG */
	    PvmUnpackStats();
	    nstat++;
	    break;

	case PVM_SLAVE_EXIT:		/* Slave has exited. */
	    Status_Info("\nSlave at %s has exited.\n",slave_name);
	    
	    if (--minions <= 0)  {
		Error_Line( "Error - All slave tasks have exited!\n");
		pvm_exit();
		return;
	    }
	    break;
	    
        case PVM_ERROR_REPORT:          /* slaves reports an error */
            PvmUnpackErrorReport();
            break;

	default:			/* something strange happend */
	    Warning(0.0, "Bad block recieved from %s - message type %4X.\n",
			slave_name, msgtag);
	}
    } /* incomplete frames */

    /* Finished calculating everything */
    PvmMasterEnd();
}

/*
 * Stop PvmMaster
 */
void PvmMasterEnd() {
    int i;
    static int already_stopped = 0;

    if (already_stopped) return;
    already_stopped=1;

    /* tell all slave to stop work */
    for(i=0; i < PvmTasks; i++) {
	pvm_kill(tids[i]); /* FD 00-10-22 */
	pvm_initsend(PvmDataDefault);
	pvm_send(tids[i], PVM_STOP_SLAVE);
    }

    /* Wait a few seconds for lingering stats */
    wait_time.tv_sec  = 4;
    wait_time.tv_usec = 0;

    if (opts.Options & VERBOSE)
	Status_Info("\nWaiting for remaining slave stats.\n");

    /* some incoming stats */
    while ((bufid = pvm_trecv(-1, PVM_SLAVE_STATS, &wait_time)) > 0)  {
#ifdef DEBUG
	int msgtag;
	if (pvm_bufinfo(bufid, &bytesin, &msgtag, &rtid) < 0) {
	    Error_Line( "PVM Master cannot bufinfo.\n");
	    pvm_perror("");
	    pvm_exit();
	    return;
	}

	PvmIdentifySlave();

	Status_Info("Getting stats from %s.\n", slave_name);
#endif /* DEBUG */

	PvmUnpackStats();
	nstat++;
	if (nstat >= PvmTasks) break; /* FD 06/00 */
    }

    /* Dump stats for the tasks */
    Statistics( "\n\nPVM Task Distribution Statistics:\n");
    Statistics( "%20s  [ done ] [ late ]","host name");
    Statistics( "%20s  [ done ] [ late ]\n","host name");

    for (i = 0; i < PvmTasks; i++) {
	int dtid;
	int j;

	dtid = pvm_tidtohost(tids[i]);

	for (j = 0; j < PvmTasks; j++)	{
	    if (dtid == hostinfo[j].hi_tid)
		break;
	}

	Statistics("%20s  [%5.2f%%] [%5.2f%%]", hostinfo[j].hi_name,
		   (DBL)slave_stat[i].pixels_done * 100 / npixels_done / frames,
		   (DBL)slave_stat[i].pixels_late * 100 / npixels_done / frames);

	if (i % 2 == 1)
	    Statistics("\n");
    }

    pvm_exit();

    Statistics("\n\nPOV-Ray statistics for finished frames:");
    if(opts.FrameSeq.FrameType==FT_MULTIPLE_FRAME) {
	for(i=0; i < next_frame_finish; i++) {
	    Statistics("\nFrame %d:\n", opts.FrameSeq.InitialFrame + i);
	    PRINT_STATS(frame_stat[i].stats);
	}
	if (opts.Do_Stats) {
	    Statistics("\n\nTotal statistics:");
	}
    }

}

/*
 * A slave sent some stats
 */
static void PvmUnpackStats() {
    COUNTER tmpStat[MaxStat];

    pvm_upkint(&frame, 1, 1);

    if( pvm_upkSTAT((long*)tmpStat, MaxStat*sizeof(COUNTER)/sizeof(long), 1) >= 0) {
	int i;
	COUNTER tmp;

	for(i=0; i < MaxStat; i++) {
	    /* stats per frame */
	    Add_Counter(tmp,tmpStat[i],frame_stat[frame].stats[i]);
	    frame_stat[frame].stats[i] = tmp;

	    /* total stats */
	    Add_Counter(tmp,tmpStat[i],totalstats[i]);
	    totalstats[i] = tmp;

	    if(opts.FrameSeq.FrameType != FT_MULTIPLE_FRAME) {
		Add_Counter(tmp,tmpStat[i],stats[i]);
		stats[i] = tmp;
	    }
	}
    } else {
	Error_Line("Unpack Stats.\n");
	pvm_perror("");
    }
}

/*
 * A slave sent some error
 */
static void PvmUnpackErrorReport() {
    char vsbuffer[1000];

    pvm_upkstr(vsbuffer);
    Error_Line("%s [%x] reported: %s\n", slave_name, rtid, vsbuffer);
}


/********************************************************************************
 * Slave Intialization
 ********************************************************************************/

int Pvm_Slave_Init() {
    /* Get master TID */
    PvmMTid = pvm_parent();

    /* Have PVM notify us if the host POV process is killed
     * or the local PVM daemon dies */
    pvm_notify(PvmTaskExit, PVM_KILL_SLAVE, 1, &PvmMTid);
    pvm_setopt(PvmAutoErr, 2);
    
    /* Set direct routing */
    pvm_setopt(PvmRoute, PvmRouteDirect);
    
    /* Tell the master we started successfully */
    pvm_initsend(PvmDataDefault);
    pvm_send(PvmMTid, PVM_INIT_SLAVE);
    
    /* Send an initial work request to the master, so that there is always
     * an outstanding block assignment, and we don't have to wait for the
     * master to respond to our request.
     */
    pvm_initsend(PvmDataDefault);
    pvm_send(PvmMTid, PVM_NEED_WORK);
    
    NICE(PvmNice);

    opts.Options |= DISKWRITE; /* FD 00-10-18 : Slaves must be able to write, since it is the way they communicate with the master */
    return 0;
}


/********************************************************************************
 * Slave Control
 ********************************************************************************/

void PvmReportError(char *vsbuffer) {
    pvm_initsend(PvmDataDefault);
    pvm_pkstr(vsbuffer);
    pvm_send(PvmMTid, PVM_ERROR_REPORT);
}

static void PvmSendStats() {
    pvm_initsend(PvmDataDefault);
    pvm_pkint( &frame, 1, 1);
    pvm_pkSTAT( (long*)stats, MaxStat*sizeof(COUNTER)/sizeof(long), 1);
    pvm_send(PvmMTid, PVM_SLAVE_STATS);
}

/*
 * Request a new block from the master,
 * handle all messages sent by the master
 * ret: 0 .. continue with new frame (or exit)
 *      1 .. continue with this frame (but new region)
 */
int Pvm_Slave_Control() {

    struct timeval wait_time;
    int bufid;
    unsigned int msgtag;
    int f;
    static int first = 1;

    /* Wait up to 5 minutes for a message, otherwise exit */
    wait_time.tv_sec  = 300;
    wait_time.tv_usec = 0;

    if((bufid = pvm_trecv(PvmMTid, -1, &wait_time)) == 0) {
	Error_Line( "Slave timed out waiting for master\n");
	pvm_exit();
	exit(1);
    } else if (bufid < 0) {
	Error_Line( "Cannot receive message from master.");
	pvm_perror("");
	pvm_exit();
	exit(1);
    }

    /* We have a block assignment from the master - IGOR SERVES! */
    if(pvm_bufinfo(bufid, (int *)0, &msgtag, &PvmMTid) < 0) {
	Error_Line("Cannot get PVM bufinfo\n");
	pvm_perror("");
	pvm_exit();
	exit(1);
    }
    
    switch( msgtag) {
    case PVM_INIT_SLAVE:
	pvm_initsend(PvmDataDefault);
	pvm_send(PvmMTid, PVM_NEED_WORK);
	break;

    case PVM_WORK:
	/* Ask master for some more block assignments in advance
	 * to avoid waiting for them to be sent later
	 */
	pvm_initsend(PvmDataDefault);
	pvm_send(PvmMTid, PVM_NEED_WORK);

	/* Unpack the message */
	pvm_upkint(&f, 1, 1);
	pvm_upkint(&PvmBlockNum, 1, 1);
	pvm_upkint(&opts.First_Line, 1, 1);
	pvm_upkint(&opts.Last_Line, 1, 1);
	pvm_upkint(&opts.First_Column, 1, 1);
	pvm_upkint(&opts.Last_Column, 1, 1);

	/* Last_Line is really "number of lines to do" */
	opts.Last_Line++;

	/* Send the data every 1/4 of a block */
	PvmRowsToSend = (opts.Last_Line - opts.First_Line + 3) / 4;

	/* Last_Column is really "number of columns to do" */
	opts.Last_Column++;
	
	/* check, if a new frame starts */
	if(frame != f) {
	    if(!first)
		PvmSendStats();
	    first = 0;
	    PvmNextFrame = opts.FrameSeq.InitialFrame + f;
	    frame = f;
	    return 0;		/* stop with this frame */
	}
	
	/* We need to do this for every block */
	Initialize_Renderer();
	
	return 1;		/* still the same frame */

    case PVM_STOP_SLAVE:	/* Everything is done, send stats and exit */
	PvmSendStats();
	PvmNextFrame = opts.FrameSeq.FinalFrame + 1;
	return 0;

    case PVM_KILL_SLAVE:
	Error_Line("\nSlave killed because of dead master.\n");
	pvm_exit();
	exit(1);

    case PVM_POLL_SLAVE:
      /* Client must ask for new work */
      wait_time.tv_sec  = 1; /* 1 sec delay for polling - small clusters */
      wait_time.tv_usec = 0;
      select(0, 0, 0, 0, &wait_time);
      pvm_initsend(PvmDataDefault);
      pvm_send(PvmMTid,PVM_NEED_WORK);  /* Request work again */
      return Pvm_Slave_Control();
      break;

    default: /* Unknown message */
	Error_Line("\nUnknown control message from master - %X\n",msgtag);
	pvm_initsend(PvmDataDefault);
	pvm_send(PvmMTid,PVM_NEED_WORK);  /* Request work again */
    }
    return 0;
}

void Pvm_Slave_Exit() {
    pvm_initsend(PvmDataDefault);
    pvm_send(PvmMTid, PVM_SLAVE_EXIT);
    pvm_exit();
}

void Pvm_Write_Line(handle, line_data, line_number)
    FILE_HANDLE     *handle;
    COLOUR          *line_data;
    int             line_number;
{
    static int in_buffer = 0;
    static int ncols;

    if (in_buffer == 0) /* No lines written yet. Initialize buffer */
    {
	ncols = opts.Last_Column - opts.First_Column;
	pvm_initsend(PvmDataDefault);

	/* Insert block info into message */
	pvm_pkint(&frame, 1, 1);
	pvm_pkint(&PvmBlockNum, 1, 1);
	pvm_pkint(&line_number, 1, 1);
	pvm_pkint(&opts.First_Column, 1, 1);
	pvm_pkint(&ncols, 1, 1);
    }
#ifdef DEBUG
    Status_Info("sending frame: %d, row: %d, col: %d+%d\n",
                frame, line_number, opts.First_Column, ncols);
#endif
    if (pvm_pkCOL(&(line_data[opts.First_Column][0]),5*ncols, 1) < 0) {
	Error_Line( "Cannot pack in Pvm_Write_Line\n");
	pvm_perror("");
	if (in_buffer > 0) {
	    pvm_send(PvmMTid, PvmBlockNum);  /* Send any lines we have done. */
	}
	pvm_initsend(PvmDataDefault);
	pvm_send(PvmMTid, PVM_SLAVE_EXIT); /* Tell the master we're dying */
	pvm_exit();
	return;
    }

    in_buffer++;
    
    if (in_buffer >= PvmRowsToSend || line_number >= opts.Last_Line - 1) {
	if (pvm_send(PvmMTid, PVM_SLAVE_RESULTS) < 0) { 
	    Error_Line( "Cannot send block in Pvm_Write_Line\n");
	    pvm_perror("");
	    pvm_exit();
	    return;
	}

	in_buffer = 0;
    }
} 



















